unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

type
  TForm1 = class(TForm)
    OpenDialog1: TOpenDialog;
    LabelFileName: TLabel;
    LabelLineCount: TLabel;
    LabelErrorCount: TLabel;
    LabelStatus: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Button1: TButton;
    Label8: TLabel;
    Label9: TLabel;
    LabelHexFile: TLabel;
    LabelListFile: TLabel;
    Label1: TLabel;
    Button2: TButton;
    Label2: TLabel;
    procedure Button1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Button2Click(Sender: TObject);
  private
    { Private declarations }
    procedure AssembleFile;
    procedure EvalWords;
    procedure Expressions;
    function  FindOpCode( S: string ) : integer;
    procedure IncludeInputFile( S: string );
    procedure InitArrays;
    function  IntToHex( I: integer ) : string;
    procedure LoadOpCodes;
    procedure LogError( S: string );
    procedure LogWarn( S: string );
    procedure LogXlate;
    function  Numericize( S: String ) : integer;
    procedure ParseWords( S: string );
    procedure WriteEqu;
    procedure WriteHex;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

{ ***************** global variables and tables ******************** }

var
   opcode_mnuemonic: array [1..45] of string;
   opcode_value: array [1..45] of integer;
   opcode_min_parms: array [1..45] of integer;
   opcode_max_parms: array [1..45] of integer;
   opcode_parm1_max: array [1..45] of integer;
   opcode_parm2_max: array [1..45] of integer;
   opcode_parm2_default: array [1..45] of integer;
   opcode_parm1_trunc: array [1..45] of integer;
   opidx: integer;

   { storage for program memory }
   pm_byte: array [0..5000] of integer;
   pm_used: array [0..5000] of boolean;
   pmidx: integer;

   { storage of equates }
   equ_word: array [1..2000] of string;
   equ_val: array [1..2000] of integer;
   equidx: integer;

   org: integer;
   line_count: integer;
   error_count: integer;

   words: array[1..40] of string;

   ListFileName: string;
   ListFile: TextFile;
   ListFileActive: boolean;
   HexFileName: string;
   HexFile: TextFile;

   XlateString: string;

   LinePrefix: string;
   PendingInclude: string;

   PassNumber: integer;
   BankOption: boolean;

   cond: array [1..10] of boolean;
   ifidx: integer;


{ ******************** custom written routines ******************* }


procedure TForm1.AssembleFile();
var
   ptr, limit: integer;
   s1, s2: string;
begin
   InitArrays();
   LabelHexFile.Caption := '';
   s1 := string(OpenDialog1.FileName);
   ptr := Length( s1 );
   limit := ptr - 5;
   if limit < 1 then
      limit := 1;
   s2 := '';
   while ptr > limit do
      begin
         if s1[ptr] = '.' then
            begin
               s2 := s1;
               SetLength( s2, ptr - 1 );
               ptr := limit;
            end
         else
            ptr := ptr - 1;
      end;
   if Length(s2) = 0 then
      s2 := s1;
   ListFileName := s2 + '.lst';
   LabelListFile.Caption := ListFileName;
   AssignFile(ListFile, ListFileName);
   Rewrite(ListFile);
   ListFileActive := true;
   LabelFileName.Caption := s1;
   WriteLn(ListFile, 'Pass 1: compute label locations');
   PassNumber := 1;
   LabelStatus.Caption := 'pass 1';
   IncludeInputFile( s1 );
   if error_count = 0 then
      begin
         WriteLn(ListFile, 'Pass 2: assemble opcodes');
         LabelStatus.Caption := 'pass 2';
         PassNumber := 2;
         IncludeInputFile( s1 );
      end;
   if error_count = 0 then
      begin
         WriteEqu();
         HexFileName := s2 + '.hex';
         LabelHexFile.Caption := HexFileName;
         AssignFile(HexFile, HexFileName);
         Rewrite(HexFile);
         LabelStatus.Caption := 'writing hex';
         WriteHex();
         CloseFile(HexFile);
         LabelStatus.Caption := 'completed';
      end
   else
      LabelStatus.Caption := 'errors found';
   CloseFile(ListFile);
end;


procedure TForm1.EvalWords;
var
   idx, idx2, idx3, idxend, trunc: integer;
   op1, op2, num1, num2, num3, bank: integer;
   p1: integer;
   s1, s2, s3: string;
   clean_pass: boolean;
begin
   if words[1] = '' then
      exit;

   if ( Uppercase(words[1]) = 'IFDEF' ) or ( Uppercase(words[1]) = 'IFNDEF' ) then
      begin
         ifidx := ifidx + 1;
         if ifidx > 10 then
            begin
               LogError('too many nested IFDEF/IFNDEF levels');
               ifidx := 10;
            end;
         s2 := words[2];
         equidx := 1;
         while ( equ_word[equidx] <> ' ' ) and ( equ_word[equidx] <> UpperCase(words[2]) ) do
            equidx := equidx + 1;
         cond[ifidx] := true;
         if ( equ_word[equidx] = ' ' ) and ( Uppercase(words[1]) = 'IFDEF' ) then
            cond[ifidx] := false;
         if ( equ_word[equidx] <> ' ' ) and ( Uppercase(words[1]) = 'IFNDEF' ) then
            cond[ifidx] := false;
         exit;
      end;

   if Uppercase(words[1]) = 'ELSE' then
      begin
         if ifidx < 1 then
            LogError('ELSE without matching IFDEF/IFNDEF')
         else
            cond[ifidx] := not cond[ifidx];
         exit;
      end;

   if Uppercase(words[1]) = 'ENDIF' then
      begin
         if ifidx < 1 then
            LogError('ENDIF without matching IFDEF/IFNDEF')
         else
            begin
               cond[ifidx] := true;
               ifidx := ifidx - 1;
            end;
         exit;
      end;


   if cond[1] and cond[2] and cond[3] and cond[4] and cond[5] and cond[6] and cond[7] and cond[8] and cond[9] and cond[10] then
      begin
      end
   else
      exit;

   if Uppercase(words[1]) = 'INCLUDE' then
      begin
         if words[2] = '' then
            LogError('expected filename after include')
         else
            if words[3] <> '' then
               LogError('too many parms for INCLUDE directive')
            else
               begin
                  s1 := words[2];
                  PendingInclude := s1;
               end;
         exit;
      end;

   if Uppercase(words[1]) = 'LIST' then
      begin
         ListFileActive := true;
         exit;
      end;

   if Uppercase(words[1]) = 'NOLIST' then
      begin
         ListFileActive := false;
         exit;
      end;

   Expressions();

   if ( Uppercase(words[2]) = 'EQU' ) then
      begin
         { numericize }
         idx := 3;
         while idx < 41 do
            begin
               if ( words[idx] = '' )
                     or ( words[idx] = '&' )
                     or ( words[idx] = '|' )
                     or ( words[idx] = '+' )
                     or ( words[idx] = '-' ) then
                  begin
                  end
               else
                  begin
                     idxend := idx;
                     num3 := Numericize(words[idx]);
                     words[idx] := '0' + IntToStr(num3);
                  end;
               idx := idx + 1;
            end;
         num3 := Numericize(words[3]);
         equidx := 1;
         while ( equ_word[equidx] <> ' ' ) and ( equ_word[equidx] <> UpperCase(words[1]) ) do
            equidx := equidx + 1;
         equ_word[equidx] := UpperCase(words[1]);
         equ_val[equidx] := num3;
         exit;
      end;

   if Uppercase(words[1]) = 'ORG' then
      begin
         if words[3] <> '' then
            LogError('too many operands for ORG directive');
         org := Numericize( words[2] );
         exit;
      end;

   if Uppercase(words[1]) = 'MEM' then
      begin
         if PassNumber = 2 then
            begin
               if words[3] <> '' then
                  LogError('too many operands for MEM directive');
               num1 := Numericize( words[2] );
               pm_byte[org] := num1;
               pm_used[org] := true;
               s1 := IntToHex(org);
               s2 := IntToHex(num1);
               LinePrefix := s1 + ':' + s2 + '      ';
            end;
         org := org + 1;
         exit;
      end;

   { search for opcode as word1 or word2 }

   op1 := FindOpCode( words[1] );
   if ( op1 = 0 ) and ( words[2] <> '' ) then
      op2 := FindOpCode( words[2] )
   else
      op2 := 0;

   if ( op1 = 0 ) and ( words[1] <> '' ) and ( PassNumber = 1 ) then
      begin
         equidx := 1;
         while ( equ_word[equidx] <> ' ' ) and ( equ_word[equidx] <> UpperCase(words[1]) ) do
            equidx := equidx + 1;
         equ_word[equidx] := UpperCase(words[1]);
         equ_val[equidx] := org;
      end;

   if PassNumber = 1 then
      begin
         if ( op1 > 0 ) or ( op2 > 0 ) then
            org := org + 1;
         exit;
      end;

   if ( op1 = 0 ) and ( words[2] = '' ) then  {just a label followed by nothing...}
      exit;

   { if you've gotten this far, then you're assembling an opcode during pass 2 }

   { if we had a label, shift everything over so they're all lined up the same }
   if op1 = 0 then
      begin
         words[1] := words[2];
         words[2] := words[3];
         words[3] := words[4];
         words[4] := '';
         words[5] := '';
         op1 := op2;
      end;

   { NEW NEW NEW ---- look for "scenix" syntax, and xlate into "microchip" syntax }

   XlateString := '';

   if op1 = 999 then { special value used to denote opcodes that must be translated }
      begin
         if Uppercase(words[1]) = 'ADD' then
            begin
               if ( Uppercase(words[2]) = 'W' ) and ( words[3] <> '' ) then   { ADD W ?  -->  ADDWF ? W }
                  begin
                     words[1] := 'ADDWF';
                     words[2] := words[3];
                     words[3] := 'W';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
               if Uppercase(words[3]) = 'W' then       { ADD ? W  -->  ADDWF ? }
                  begin
                     words[1] := 'ADDWF';
                     words[3] := '';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
                  begin
                     LogError('expected pattern "ADD W ?" or "ADD ? W"');
                     exit;
                  end;
            end
         else
         if Uppercase(words[1]) = 'AND' then
            begin
               if ( Uppercase(words[2]) = 'W' ) and ( copy(words[3],1,1) = '#' ) then   { AND W #?  -->  ANDLW ? }
                  begin
                     words[1] := 'ANDLW';
                     words[2] := copy(words[3], 2, length(words[3]) - 1 );
                     words[3] := '';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
               if ( Uppercase(words[2]) = 'W' ) and ( words[3] <> '' ) then   { AND W ?  -->  ANDWF ? W }
                  begin
                     words[1] := 'ANDWF';
                     words[2] := words[3];
                     words[3] := 'W';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
               if Uppercase(words[3]) = 'W' then   { AND ? W  -->  ANDWF ? }
                  begin
                     words[1] := 'ANDWF';
                     words[3] := '';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
                  begin
                     LogError('expected pattern "AND W ?", "AND ? W" or "AND W #?"');
                     exit;
                  end;
            end
         else
         if Uppercase(words[1]) = 'CLC' then
            begin
               words[1] := 'BCF';
               words[2] := 'STATUS';
               words[3] := 'C';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'CLR' then
            begin
               if Uppercase(words[2]) = 'W' then   { CLR W  -->  CLRW }
                  begin
                     words[1] := 'CLRW';
                     words[2] := '';
                     words[3] := '';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
               if Uppercase(words[2]) = 'WDT' then   { CLR WDT  -->  CLRWDT }
                  begin
                     words[1] := 'CLRWDT';
                     words[2] := '';
                     words[3] := '';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
                  begin                            { CLR ?  -->  CLRF ? }
                     words[1] := 'CLRF';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end;
            end
         else
         if Uppercase(words[1]) = 'CLRB' then
            begin
               words[1] := 'BCF';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'CLZ' then
            begin
               words[1] := 'BCF';
               words[2] := 'STATUS';
               words[3] := 'Z';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'DEC' then
            begin
               words[1] := 'DECF';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'DECSZ' then
            begin
               words[1] := 'DECFSZ';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'INC' then
            begin
               words[1] := 'INCF';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'INCSZ' then
            begin
               words[1] := 'INCFSZ';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'JMP' then
            begin
               if Uppercase(words[2]) = 'PC+W' then
                  begin
                     words[1] := 'ADDWF';
                     words[2] := 'PCL';
                     words[3] := '';
                  end
               else
               if Uppercase(words[2]) = 'W' then
                  begin
                     words[1] := 'MOVWF';
                     words[2] := 'PCL';
                     words[3] := '';
                  end
               else
                  begin
                     words[1] := 'GOTO';
                  end;
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'MOV' then
            begin
               if Uppercase(words[2]) = 'W' then
                  begin
                     if ( Uppercase(words[3]) = 'M' ) or ( Uppercase(words[3]) = '!MODE' ) then    { MOV W M  -->  MOVMW }
                        begin
                           words[1] := 'MOVMW';
                           words[2] := '';
                           words[3] := '';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                     if copy(words[3], 1, 1) = '#' then   { MOV W #?  -->  MOVLW ? }
                        begin
                           words[1] := 'MOVLW';
                           words[2] := copy(words[3], 2, length(words[3]) - 1);
                           words[3] := '';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                     if copy(words[3], 1, 1) = '/' then   { MOV W /?  -->  NOTF ? W }
                        begin
                           words[1] := 'NOTF';
                           words[2] := copy(words[3], 2, length(words[3]) - 1);
                           words[3] := 'W';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                     if copy(words[3], 1, 2) = '--' then   { MOV W --?  -->  DECF ? W }
                        begin
                           words[1] := 'DECF';
                           words[2] := copy(words[3], 3, length(words[3]) - 2);
                           words[3] := 'W';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                     if copy(words[3], 1, 2) = '++' then   { MOV W ++?  -->  INCF ? W }
                        begin
                           words[1] := 'INCF';
                           words[2] := copy(words[3], 3, length(words[3]) - 2);
                           words[3] := 'W';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                     if copy(words[3], 1, 2) = '<<' then   { MOV W <<?  -->  RLF ? W }
                        begin
                           words[1] := 'RLF';
                           words[2] := copy(words[3], 3, length(words[3]) - 2);
                           words[3] := 'W';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                     if copy(words[3], 1, 2) = '>>' then   { MOV W >>?  -->  RRF ? W }
                        begin
                           words[1] := 'RRF';
                           words[2] := copy(words[3], 3, length(words[3]) - 2);
                           words[3] := 'W';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                     if copy(words[3], 1, 2) = '<>' then   { MOV W <>?  -->  SWAPF ? W }
                        begin
                           words[1] := 'SWAPF';
                           words[2] := copy(words[3], 3, length(words[3]) - 2);
                           words[3] := 'W';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                     if Uppercase(copy(words[3], length(words[3]) - 1, 2)) = '-W' then   { MOV W ?-W  -->  SUBWF ? W }
                        begin
                           words[1] := 'SUBWF';
                           words[2] := copy(words[3], 1, length(words[3]) - 2);
                           words[3] := 'W';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                        begin   { MOV W ?  -->  MOVFW ? }
                           words[1] := 'MOVFW';
                           words[2] := words[3];
                           words[3] := '';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end;
                  end
               else
                  begin
                     if ( Uppercase(words[2]) = 'M' ) and ( Uppercase(words[3]) = 'W' ) then   { MOV M W  -->  MOVWM }
                        begin
                           words[1] := 'MOVWM';
                           words[2] := '';
                           words[3] := '';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                     if ( copy(words[2], 1, 1) = '!' ) and ( Uppercase(words[3]) = 'W' ) then
                        begin
                           if Uppercase(words[2]) = '!MODE' then
                              begin
                                 words[1] := 'MOVWM';
                                 words[2] := '';
                                 words[3] := '';
                                 op1 := FindOpCode(words[1]);
                                 LogXlate();
                              end
                           else
                           if Uppercase(words[2]) = '!OPTION' then
                              begin
                                 words[1] := 'MOVWO';
                                 words[2] := '';
                                 words[3] := '';
                                 op1 := FindOpCode(words[1]);
                                 LogXlate();
                              end
                           else
                           if Uppercase(words[2]) = '!RA' then
                              begin
                                 words[1] := 'TRIS';
                                 words[2] := '5';
                                 words[3] := '';
                                 op1 := FindOpCode(words[1]);
                                 LogXlate();
                              end
                           else
                           if Uppercase(words[2]) = '!RB' then
                              begin
                                 words[1] := 'TRIS';
                                 words[2] := '6';
                                 words[3] := '';
                                 op1 := FindOpCode(words[1]);
                                 LogXlate();
                              end
                           else
                           if Uppercase(words[2]) = '!RC' then
                              begin
                                 words[1] := 'TRIS';
                                 words[2] := '7';
                                 words[3] := '';
                                 op1 := FindOpCode(words[1]);
                                 LogXlate();
                              end
                           else
                              begin
                                 LogError('exclamation may be used for "!RA", "!RB", "!RC", "!OPTION" and "!MODE" only');
                                 exit;
                              end;
                        end
                     else
                     if ( Uppercase(words[2]) = 'M' ) and ( copy(words[3], 1, 1) = '#' ) then  { MOV M #?  -->  MODE ? }
                        begin
                           words[1] := 'MODE';
                           words[2] := copy(words[3], 2, length(words[3]) - 1);
                           words[3] := '';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                     if Uppercase(words[3]) = 'W' then   { MOV ? W  -->  MOVWF ? }
                        begin
                           words[1] := 'MOVWF';
                           words[3] := '';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                        begin
                           LogError('invalid MOV syntax');
                           exit;
                        end;
                  end;
            end
         else
         if Uppercase(words[1]) = 'MOVSZ' then
            begin
               if Uppercase(words[2]) = 'W' then
                  begin
                     if copy(words[3], 1, 2) = '--' then  { MOVSZ W --?  -->  DECF ? W }
                        begin
                           words[1] := 'DECFSZ';
                           words[2] := copy(words[3], 3, length(words[3]) - 2);
                           words[3] := 'W';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                     if copy(words[3], 1, 2) = '++' then  { MOVSZ W ++?  -->  INCF ? W }
                        begin
                           words[1] := 'INCFSZ';
                           words[2] := copy(words[3], 3, length(words[3]) - 2);
                           words[3] := 'W';
                           op1 := FindOpCode(words[1]);
                           LogXlate();
                        end
                     else
                        begin
                           LogError('expected pattern "MOVSZ W --?" or "MOVSZ W ++?"');
                           exit;
                        end;
                  end
               else
                  begin
                     LogError('expected pattern "MOVSZ W --?" or "MOVSZ W ++?"');
                     exit;
                  end;
            end
         else
         if Uppercase(words[1]) = 'NOT' then
            begin
               if Uppercase(words[2]) = 'W' then
                  begin
                     words[1] := 'XORLW';
                     words[2] := '$FF';
                     words[3] := '';
                  end
               else
                  begin
                     words[1] := 'NOTF';
                  end;
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'OR' then
            begin
               if ( Uppercase(words[2]) = 'W' ) and ( copy(words[3],1,1) = '#' ) then   { OR W #?  -->  IORLW ? }
                  begin
                     words[1] := 'IORLW';
                     words[2] := copy(words[3],2,length(words[3]) - 1);
                     words[3] := '';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
               if ( Uppercase(words[2]) = 'W' ) and ( words[3] <> '' ) then   { OR W ?  -->  IORWF ? W }
                  begin
                     words[1] := 'IORWF';
                     words[2] := words[3];
                     words[3] := 'W';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
               if Uppercase(words[3]) = 'W' then   { OR ? W  -->  IORWF ? }
                  begin
                     words[1] := 'IORWF';
                     words[3] := '';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
                  begin
                     LogError('expected pattern "OR W ?", "OR ? W" or "OR W #?"');
                     exit;
                  end;
            end
         else
         if Uppercase(words[1]) = 'RET' then
            begin
               words[1] := 'RETURN';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'RL' then
            begin
               words[1] := 'RLF';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'RR' then
            begin
               words[1] := 'RRF';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'SB' then
            begin
               words[1] := 'BTFSS';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'SC' then
            begin
               words[1] := 'BTFSS';
               words[2] := 'STATUS';
               words[3] := 'C';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'SETB' then
            begin
               words[1] := 'BSF';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'SKIP' then
            begin
               if ( org mod 2 ) = 0 then
                  begin
                     words[1] := 'BTFSC';
                     words[2] := 'PCL';
                     words[3] := '7';
                  end
               else
                  begin
                     words[1] := 'BTFSS';
                     words[2] := 'PCL';
                     words[3] := '7';
                  end;
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'SNB' then
            begin
               words[1] := 'BTFSC';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'SUB' then
            begin
               if ( Uppercase(words[2]) = 'W' ) and ( words[3] <> '' ) then   { SUB W ?  -->  SUBWF ? W }
                  begin
                     words[1] := 'SUBWF';
                     words[2] := words[3];
                     words[3] := 'W';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
               if Uppercase(words[3]) = 'W' then       { SUB ? W  -->  SUBWF ? }
                  begin
                     words[1] := 'SUBWF';
                     words[3] := '';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
                  begin
                     LogError('expected pattern "SUB W ?" or "SUB ? W"');
                     exit;
                  end;
            end
         else
         if Uppercase(words[1]) = 'SWAP' then
            begin
               words[1] := 'SWAPF';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'TEST' then
            begin
               words[1] := 'TESTF';
               op1 := FindOpCode(words[1]);
               LogXlate();
            end
         else
         if Uppercase(words[1]) = 'XOR' then
            begin
               if ( Uppercase(words[2]) = 'W' ) and ( copy(words[3],1,1) = '#' ) then   { XOR W #?  -->  XORLW ? }
                  begin
                     words[1] := 'XORLW';
                     words[2] := copy(words[3], 2, length(words[3]) - 1 );
                     words[3] := '';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
               if ( Uppercase(words[2]) = 'W' ) and ( words[3] <> '' ) then   { XOR W ?  -->  XORWF ? W }
                  begin
                     words[1] := 'XORWF';
                     words[2] := words[3];
                     words[3] := 'W';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
               if Uppercase(words[3]) = 'W' then   { XOR ? W  -->  XORWF ? }
                  begin
                     words[1] := 'XORWF';
                     words[3] := '';
                     op1 := FindOpCode(words[1]);
                     LogXlate();
                  end
               else
                  begin
                     LogError('expected pattern "XOR W ?", "XOR ? W" or "XOR W #?"');
                     exit;
                  end;
            end;
         if op1 = 999 then
            begin
               LogError('Translation failed');
               exit;
            end;
      end;

   { end of translations from scenix to microchip syntax }

   if op1 = 0 then
      begin
         LogError('Expected OpCode as first or second word');
         exit;
      end;

   { check number of parameters against table }
   idx := 5;
   while words[idx] = '' do
      idx := idx - 1;
   idx := idx - 1;

   if idx < opcode_min_parms[op1] then
      begin
         LogError('too few parameters');
         exit;
      end;

   if idx > opcode_max_parms[op1] then
      begin
         LogError('too many parameters');
         exit;
      end;

   if words[2] <> '' then
      num1 := Numericize( words[2] )
   else
      num1 := 0;
   if words[3] <> '' then
      num2 := Numericize( words[3] )
   else
      num2 := opcode_parm2_default[op1];

   if num1 > opcode_parm1_max[op1] then
      begin
         LogError('parm 1 exceeds maximum value');
         exit;
      end;

   if num2 > opcode_parm2_max[op1] then
      begin
         LogError('parm 2 exceeds maximum value');
         exit;
      end;

   trunc := opcode_parm1_trunc[op1];
   if num1 >= trunc then
      begin
         num3 := num1 div trunc;
         num1 := num1 mod trunc;
         bank := num3 * trunc;
         if BankOption then
            begin {before you add this code... realize it effects Label addresses... }
            end
         else
            begin
               if ( bank > 0 ) then
                  LogWarn('BANK or PAGE other than 0 assumed');
               if ( bank > 0 ) and ( num1 < 16 ) and ( trunc = 32 ) then
                  LogWarn('Lower-half Register specified outside bank 0');
            end;
      end;

   { special processing for unusual or restrictive opcodes... }

   if opcode_mnuemonic[op1] = 'TRIS' then
      begin
         if ( num1 < 5 ) or ( num1 > 7 ) then
            LogError('expected value of 5, 6 or 7 only');
      end
   else
   if opcode_mnuemonic[op1] = 'BANK' then
      begin
         if ( ( num1 mod 32 ) <> 0 ) and ( num1 > 7 ) then
            begin
               LogWarn('Bank value not on proper boundary');
            end;
         if num1 > 7 then
            num1 := num1 div 32;
      end
   else
   if opcode_mnuemonic[op1] = 'PAGE' then
      begin
         if ( ( num1 mod 512 ) <> 0 ) and ( num1 > 7 ) then
            begin
               LogWarn('Page value not on proper boundary');
            end;
         if num1 > 7 then
            num1 := num1 div 512;
         if num1 > 3 then
            LogWarn('current devices have 2048 words PM, using pages 0-3 (512 bytes per page)');
      end;

   { everything checks out... plug the memory }

   num3 := opcode_value[op1] or num1 or ( num2 * 32 );
   pm_byte[org] := num3;
   pm_used[org] := true;
   s1 := IntToHex(org);
   s2 := IntToHex(num3);
   LinePrefix := s1 + ':' + s2 + '      ';
   org := org + 1;
end;


procedure TForm1.Expressions;
var
   idx, idx2, idx3, idxend, trunc: integer;
   op1, op2, num1, num2, num3, bank: integer;
   s1, s2, s3: string;
   clean_pass: boolean;
begin
   clean_pass := false;
   while clean_pass = false do
      begin
         { assume it's a clean pass until you resolve something }
         clean_pass := true;
         { scrunch everything down }
         idx := 1;
         idx2 := 1;
         while idx2 <= 40 do
            begin
               if words[idx2] <> '' then
                  begin
                     if idx <> idx2 then
                        words[idx] := words[idx2];
                     idx := idx + 1;
                  end;
               idx2 := idx2 + 1;
            end;
         while idx <= 40 do
            begin
               words[idx] := '';
               idx := idx + 1;
            end;
         { make one pass through the words, looking for stuff to resolve together }
         idx := 1;
         idx2 := 2;
         idx3 := 3;
         while idx3 <= 40 do
            begin
               s1 := words[idx];
               s2 := words[idx2];
               s3 := words[idx3];
               if ( s2 = '&' )
                     or ( s2 = '|' )
                     or ( s2 = '+' )
                     or ( s2 = '-' ) then
                  begin
                     if ( s1 <> '' ) and ( s1 <> '(' ) and ( s1 <> ')' )
                           and ( s3 <> '' ) and ( s3 <> '(' ) and ( s3 <> ')' ) then
                        begin
                           {  WriteLn( ListFile, 'expr: ' + words[idx] + ' ' + words[idx2] + ' ' + words[idx3] ); }
                           num1 := Numericize( s1 );
                           num3 := Numericize( s3 );
                           if s2 = '&' then
                              num2 := num1 and num3
                           else
                           if s2 = '|' then
                              num2 := num1 or num3
                           else
                           if s2 = '+' then
                              num2 := num1 + num3
                           else
                           if s2 = '-' then
                              num2 := num1 - num3;
                           words[idx3] := IntToStr(num2);
                           words[idx] := '';
                           words[idx2] := '';
                           clean_pass := false;
                        end
                  end
               else
               if ( s1 = '(' ) and ( s3 = ')' ) and ( s2 <> '(' ) and ( s2 <> ')' ) then
                  begin
                     words[idx] := '';
                     words[idx3] := words[idx2];
                     words[idx2] := '';
                     clean_pass := false;
                  end;
               idx := idx + 1;
               idx2 := idx2 + 1;
               idx3 := idx3 + 1;
            end;
      end;
end;



function TForm1.FindOpCode( S : string ) : integer;
var
   idx, op: integer;
   teststring: string;
begin
   teststring := Uppercase(S);
   idx := 1;
   op := 0;
   while idx <= opidx do
      begin
         if opcode_mnuemonic[idx] = teststring then
            op := idx;
         idx := idx + 1;
      end;
   if op = 0 then   { these are valid opcodes, but use special value to trigger translation }
      if       ( teststring = 'ADD' )
            or ( teststring = 'AND' )
            or ( teststring = 'CLC' )
            or ( teststring = 'CLR' )
            or ( teststring = 'CLRB' )
            or ( teststring = 'CLZ' )
            or ( teststring = 'DEC' )
            or ( teststring = 'DECSZ' )
            or ( teststring = 'INC' )
            or ( teststring = 'INCSZ' )
            or ( teststring = 'JMP' )
            or ( teststring = 'MOV' )
            or ( teststring = 'MOVSZ' )
            or ( teststring = 'NOT' )
            or ( teststring = 'OR' )
            or ( teststring = 'RET' )
            or ( teststring = 'RL' )
            or ( teststring = 'RR' )
            or ( teststring = 'SB' )
            or ( teststring = 'SC' )
            or ( teststring = 'SETB' )
            or ( teststring = 'SKIP' )
            or ( teststring = 'SNB' )
            or ( teststring = 'SUB' )
            or ( teststring = 'SWAP' )
            or ( teststring = 'TEST' )
            or ( teststring = 'XOR' ) then
         op := 999;
   FindOpCode := op;
end;


procedure TForm1.IncludeInputFile( S : string );
var
   F: TextFile;
   input, include: string;
begin
   AssignFile(F, S);
   Reset(F);
   while not Eof(F) do
      begin
         Readln(F, input);
         LinePrefix := '               ';
         ParseWords(input);
         EvalWords();
         if ( PassNumber = 2 ) or ( error_count > 0 ) and ListFileActive then
            WriteLn(ListFile, LinePrefix + input);
         if PassNumber = 2 then
            begin
               line_count := line_count + 1;
               LabelLineCount.Caption := IntToStr( line_count );
            end;
         if PendingInclude <> '' then
            begin
               include := pendinginclude;
               PendingInclude := '';
               IncludeInputFile( include );
            end;
      end;
   CloseFile(F);
end;


procedure TForm1.InitArrays;
begin
   pmidx := 0;
   while pmidx <= 5000 do
      begin
         pm_byte[pmidx] := 0;
         pm_used[pmidx] := false;
         pmidx := pmidx + 1;
      end;

   equidx := 1;
   while equidx <= 2000 do
      begin
         equ_word[equidx] := ' ';
         equ_val[equidx] := 0;
         equidx := equidx + 1;
      end;

   ifidx := 10;
   while ifidx > 0 do
      begin
         cond[ifidx] := true;
         ifidx := ifidx - 1;
      end;
   org := 0;
   line_count := 0;
   error_count := 0;
   LabelLineCount.Caption := '0';
   LabelErrorCount.Caption := '0';
   pmidx := 0;
   equidx := 1;
   BankOption := false;
end;


function TForm1.IntToHex( I: integer ) : string;
var
   d1, d2, d3, d4, amt: integer;
   s1, s2, s3, s4, hex: string;
begin
   if ( I < 0 ) or ( I > 65536 ) then
      amt := 0
   else
      amt := I;

   hex := '0123456789ABCDEF';
   d4 := amt mod 16;
   amt := amt div 16;
   d3 := amt mod 16;
   amt := amt div 16;
   d2 := amt mod 16;
   d1 := amt div 16;
   s1 := hex[d1 + 1];
   s2 := hex[d2 + 1];
   s3 := hex[d3 + 1];
   s4 := hex[d4 + 1];
   IntToHex := s1 + s2 + s3 + s4;
end;


procedure TForm1.LoadOpCodes;
begin
   opcode_mnuemonic[1]  := 'ADDWF';   opcode_value[1]  := $1C0;  opcode_min_parms[1]  := 1;  opcode_max_parms[1]  := 2;  opcode_parm1_max[1]  :=  255;  opcode_parm2_max[1]  := 1;  opcode_parm2_default[1]  := 1;  opcode_parm1_trunc[1]  := 32;
   opcode_mnuemonic[2]  := 'ANDLW';   opcode_value[2]  := $E00;  opcode_min_parms[2]  := 1;  opcode_max_parms[2]  := 1;  opcode_parm1_max[2]  :=  255;  opcode_parm2_max[2]  := 0;  opcode_parm2_default[2]  := 0;  opcode_parm1_trunc[2]  := 65536;
   opcode_mnuemonic[3]  := 'ANDWF';   opcode_value[3]  := $140;  opcode_min_parms[3]  := 1;  opcode_max_parms[3]  := 2;  opcode_parm1_max[3]  :=  255;  opcode_parm2_max[3]  := 1;  opcode_parm2_default[3]  := 1;  opcode_parm1_trunc[3]  := 32;
   opcode_mnuemonic[4]  := 'BANK';    opcode_value[4]  := $018;  opcode_min_parms[4]  := 1;  opcode_max_parms[4]  := 1;  opcode_parm1_max[4]  :=  255;  opcode_parm2_max[4]  := 0;  opcode_parm2_default[4]  := 0;  opcode_parm1_trunc[4]  := 65536;
   opcode_mnuemonic[5]  := 'BCF';     opcode_value[5]  := $400;  opcode_min_parms[5]  := 2;  opcode_max_parms[5]  := 2;  opcode_parm1_max[5]  :=  255;  opcode_parm2_max[5]  := 7;  opcode_parm2_default[5]  := 0;  opcode_parm1_trunc[5]  := 32;
   opcode_mnuemonic[6]  := 'BSF';     opcode_value[6]  := $500;  opcode_min_parms[6]  := 2;  opcode_max_parms[6]  := 2;  opcode_parm1_max[6]  :=  255;  opcode_parm2_max[6]  := 7;  opcode_parm2_default[6]  := 0;  opcode_parm1_trunc[6]  := 32;
   opcode_mnuemonic[7]  := 'BTFSC';   opcode_value[7]  := $600;  opcode_min_parms[7]  := 2;  opcode_max_parms[7]  := 2;  opcode_parm1_max[7]  :=  255;  opcode_parm2_max[7]  := 7;  opcode_parm2_default[7]  := 0;  opcode_parm1_trunc[7]  := 32;
   opcode_mnuemonic[8]  := 'BTFSS';   opcode_value[8]  := $700;  opcode_min_parms[8]  := 2;  opcode_max_parms[8]  := 2;  opcode_parm1_max[8]  :=  255;  opcode_parm2_max[8]  := 7;  opcode_parm2_default[8]  := 0;  opcode_parm1_trunc[8]  := 32;
   opcode_mnuemonic[9]  := 'CALL';    opcode_value[9]  := $900;  opcode_min_parms[9]  := 1;  opcode_max_parms[9]  := 1;  opcode_parm1_max[9]  :=  255;  opcode_parm2_max[9]  := 0;  opcode_parm2_default[9]  := 0;  opcode_parm1_trunc[9]  := 256;
   opcode_mnuemonic[10] := 'CLRF';    opcode_value[10] := $060;  opcode_min_parms[10] := 1;  opcode_max_parms[10] := 1;  opcode_parm1_max[10] :=  255;  opcode_parm2_max[10] := 0;  opcode_parm2_default[10] := 0;  opcode_parm1_trunc[10] := 32;
   opcode_mnuemonic[11] := 'CLRW';    opcode_value[11] := $040;  opcode_min_parms[11] := 0;  opcode_max_parms[11] := 0;  opcode_parm1_max[11] :=    0;  opcode_parm2_max[11] := 0;  opcode_parm2_default[11] := 0;  opcode_parm1_trunc[11] := 65536;
   opcode_mnuemonic[12] := 'CLRWDT';  opcode_value[12] := $004;  opcode_min_parms[12] := 0;  opcode_max_parms[12] := 0;  opcode_parm1_max[12] :=    0;  opcode_parm2_max[12] := 0;  opcode_parm2_default[12] := 0;  opcode_parm1_trunc[12] := 65536;
   opcode_mnuemonic[13] := 'DECF';    opcode_value[13] := $0C0;  opcode_min_parms[13] := 1;  opcode_max_parms[13] := 2;  opcode_parm1_max[13] :=  255;  opcode_parm2_max[13] := 1;  opcode_parm2_default[13] := 1;  opcode_parm1_trunc[13] := 32;
   opcode_mnuemonic[14] := 'DECFSZ';  opcode_value[14] := $2C0;  opcode_min_parms[14] := 1;  opcode_max_parms[14] := 2;  opcode_parm1_max[14] :=  255;  opcode_parm2_max[14] := 1;  opcode_parm2_default[14] := 1;  opcode_parm1_trunc[14] := 32;
   opcode_mnuemonic[15] := 'GOTO';    opcode_value[15] := $A00;  opcode_min_parms[15] := 1;  opcode_max_parms[15] := 1;  opcode_parm1_max[15] :=  511;  opcode_parm2_max[15] := 0;  opcode_parm2_default[15] := 0;  opcode_parm1_trunc[15] := 512;
   opcode_mnuemonic[16] := 'INCF';    opcode_value[16] := $280;  opcode_min_parms[16] := 1;  opcode_max_parms[16] := 2;  opcode_parm1_max[16] :=  255;  opcode_parm2_max[16] := 1;  opcode_parm2_default[16] := 1;  opcode_parm1_trunc[16] := 32;
   opcode_mnuemonic[17] := 'INCFSZ';  opcode_value[17] := $3C0;  opcode_min_parms[17] := 1;  opcode_max_parms[17] := 2;  opcode_parm1_max[17] :=  255;  opcode_parm2_max[17] := 1;  opcode_parm2_default[17] := 1;  opcode_parm1_trunc[17] := 32;
   opcode_mnuemonic[18] := 'IORLW';   opcode_value[18] := $D00;  opcode_min_parms[18] := 1;  opcode_max_parms[18] := 1;  opcode_parm1_max[18] :=  255;  opcode_parm2_max[18] := 0;  opcode_parm2_default[18] := 0;  opcode_parm1_trunc[18] := 65536;
   opcode_mnuemonic[19] := 'IORWF';   opcode_value[19] := $100;  opcode_min_parms[19] := 1;  opcode_max_parms[19] := 2;  opcode_parm1_max[19] :=  255;  opcode_parm2_max[19] := 1;  opcode_parm2_default[19] := 1;  opcode_parm1_trunc[19] := 32;
   opcode_mnuemonic[20] := 'IREAD';   opcode_value[20] := $041;  opcode_min_parms[20] := 0;  opcode_max_parms[20] := 0;  opcode_parm1_max[20] :=    0;  opcode_parm2_max[20] := 0;  opcode_parm2_default[20] := 0;  opcode_parm1_trunc[20] := 65536;
   opcode_mnuemonic[21] := 'MODE';    opcode_value[21] := $050;  opcode_min_parms[21] := 1;  opcode_max_parms[21] := 1;  opcode_parm1_max[21] :=   15;  opcode_parm2_max[21] := 0;  opcode_parm2_default[21] := 0;  opcode_parm1_trunc[21] := 65536;
   opcode_mnuemonic[22] := 'MOVFW';   opcode_value[22] := $200;  opcode_min_parms[22] := 1;  opcode_max_parms[22] := 1;  opcode_parm1_max[22] :=  255;  opcode_parm2_max[22] := 0;  opcode_parm2_default[22] := 0;  opcode_parm1_trunc[22] := 32;
   opcode_mnuemonic[23] := 'MOVLW';   opcode_value[23] := $C00;  opcode_min_parms[23] := 1;  opcode_max_parms[23] := 1;  opcode_parm1_max[23] :=  255;  opcode_parm2_max[23] := 0;  opcode_parm2_default[23] := 0;  opcode_parm1_trunc[23] := 65536;
   opcode_mnuemonic[24] := 'MOVWM';   opcode_value[24] := $043;  opcode_min_parms[24] := 0;  opcode_max_parms[24] := 0;  opcode_parm1_max[24] :=    0;  opcode_parm2_max[24] := 0;  opcode_parm2_default[24] := 0;  opcode_parm1_trunc[24] := 65536;
   opcode_mnuemonic[25] := 'MOVWF';   opcode_value[25] := $020;  opcode_min_parms[25] := 1;  opcode_max_parms[25] := 1;  opcode_parm1_max[25] :=  255;  opcode_parm2_max[25] := 0;  opcode_parm2_default[25] := 0;  opcode_parm1_trunc[25] := 32;
   opcode_mnuemonic[26] := 'MOVMW';   opcode_value[26] := $042;  opcode_min_parms[26] := 0;  opcode_max_parms[26] := 0;  opcode_parm1_max[26] :=    0;  opcode_parm2_max[26] := 0;  opcode_parm2_default[26] := 0;  opcode_parm1_trunc[26] := 65536;
   opcode_mnuemonic[27] := 'MOVWO';   opcode_value[27] := $002;  opcode_min_parms[27] := 0;  opcode_max_parms[27] := 0;  opcode_parm1_max[27] :=    0;  opcode_parm2_max[27] := 0;  opcode_parm2_default[27] := 0;  opcode_parm1_trunc[27] := 65536;
   opcode_mnuemonic[28] := 'NOP';     opcode_value[28] := $000;  opcode_min_parms[28] := 0;  opcode_max_parms[28] := 0;  opcode_parm1_max[28] :=    0;  opcode_parm2_max[28] := 0;  opcode_parm2_default[28] := 0;  opcode_parm1_trunc[28] := 65536;
   opcode_mnuemonic[29] := 'NOTF';    opcode_value[29] := $240;  opcode_min_parms[29] := 1;  opcode_max_parms[29] := 2;  opcode_parm1_max[29] :=  255;  opcode_parm2_max[29] := 1;  opcode_parm2_default[29] := 1;  opcode_parm1_trunc[29] := 32;
   opcode_mnuemonic[30] := 'PAGE';    opcode_value[30] := $010;  opcode_min_parms[30] := 1;  opcode_max_parms[30] := 1;  opcode_parm1_max[30] := 4095;  opcode_parm2_max[30] := 0;  opcode_parm2_default[30] := 0;  opcode_parm1_trunc[30] := 65536;
   opcode_mnuemonic[31] := 'RETURN';  opcode_value[31] := $00C;  opcode_min_parms[31] := 0;  opcode_max_parms[31] := 0;  opcode_parm1_max[31] :=    0;  opcode_parm2_max[31] := 0;  opcode_parm2_default[31] := 0;  opcode_parm1_trunc[31] := 65536;
   opcode_mnuemonic[32] := 'RETI';    opcode_value[32] := $00E;  opcode_min_parms[32] := 0;  opcode_max_parms[32] := 0;  opcode_parm1_max[32] :=    0;  opcode_parm2_max[32] := 0;  opcode_parm2_default[32] := 0;  opcode_parm1_trunc[32] := 65536;
   opcode_mnuemonic[33] := 'RETIW';   opcode_value[33] := $00F;  opcode_min_parms[33] := 0;  opcode_max_parms[33] := 0;  opcode_parm1_max[33] :=    0;  opcode_parm2_max[33] := 0;  opcode_parm2_default[33] := 0;  opcode_parm1_trunc[33] := 65536;
   opcode_mnuemonic[34] := 'RETP';    opcode_value[34] := $00D;  opcode_min_parms[34] := 0;  opcode_max_parms[34] := 0;  opcode_parm1_max[34] :=    0;  opcode_parm2_max[34] := 0;  opcode_parm2_default[34] := 0;  opcode_parm1_trunc[34] := 65536;
   opcode_mnuemonic[35] := 'RETLW';   opcode_value[35] := $800;  opcode_min_parms[35] := 1;  opcode_max_parms[35] := 1;  opcode_parm1_max[35] :=  255;  opcode_parm2_max[35] := 0;  opcode_parm2_default[35] := 0;  opcode_parm1_trunc[35] := 65536;
   opcode_mnuemonic[36] := 'RLF';     opcode_value[36] := $340;  opcode_min_parms[36] := 1;  opcode_max_parms[36] := 2;  opcode_parm1_max[36] :=  255;  opcode_parm2_max[36] := 1;  opcode_parm2_default[36] := 1;  opcode_parm1_trunc[36] := 32;
   opcode_mnuemonic[37] := 'RRF';     opcode_value[37] := $300;  opcode_min_parms[37] := 1;  opcode_max_parms[37] := 2;  opcode_parm1_max[37] :=  255;  opcode_parm2_max[37] := 1;  opcode_parm2_default[37] := 1;  opcode_parm1_trunc[37] := 32;
   opcode_mnuemonic[38] := 'SLEEP';   opcode_value[38] := $003;  opcode_min_parms[38] := 0;  opcode_max_parms[38] := 0;  opcode_parm1_max[38] :=    0;  opcode_parm2_max[38] := 0;  opcode_parm2_default[38] := 0;  opcode_parm1_trunc[38] := 65536;
   opcode_mnuemonic[39] := 'SUBWF';   opcode_value[39] := $080;  opcode_min_parms[39] := 1;  opcode_max_parms[39] := 2;  opcode_parm1_max[39] :=  255;  opcode_parm2_max[39] := 1;  opcode_parm2_default[39] := 1;  opcode_parm1_trunc[39] := 32;
   opcode_mnuemonic[40] := 'SWAPF';   opcode_value[40] := $380;  opcode_min_parms[40] := 1;  opcode_max_parms[40] := 2;  opcode_parm1_max[40] :=  255;  opcode_parm2_max[40] := 1;  opcode_parm2_default[40] := 1;  opcode_parm1_trunc[40] := 32;
   opcode_mnuemonic[41] := 'TESTF';   opcode_value[41] := $220;  opcode_min_parms[41] := 1;  opcode_max_parms[41] := 1;  opcode_parm1_max[41] :=  255;  opcode_parm2_max[41] := 0;  opcode_parm2_default[41] := 0;  opcode_parm1_trunc[41] := 32;
   opcode_mnuemonic[42] := 'TRIS';    opcode_value[42] := $000;  opcode_min_parms[42] := 1;  opcode_max_parms[42] := 1;  opcode_parm1_max[42] :=    7;  opcode_parm2_max[42] := 0;  opcode_parm2_default[42] := 0;  opcode_parm1_trunc[42] := 65536;
   opcode_mnuemonic[43] := 'XORLW';   opcode_value[43] := $F00;  opcode_min_parms[43] := 1;  opcode_max_parms[43] := 1;  opcode_parm1_max[43] :=  255;  opcode_parm2_max[43] := 0;  opcode_parm2_default[43] := 0;  opcode_parm1_trunc[43] := 65536;
   opcode_mnuemonic[44] := 'XORWF';   opcode_value[44] := $180;  opcode_min_parms[44] := 1;  opcode_max_parms[44] := 2;  opcode_parm1_max[44] :=  255;  opcode_parm2_max[44] := 1;  opcode_parm2_default[44] := 1;  opcode_parm1_trunc[44] := 32;

   opidx := 44;
end;


procedure TForm1.LogError( S : string );
begin
   error_count := error_count + 1;
   LabelLineCount.Caption := IntToStr( error_count );
   if XlateString <> '' then
      WriteLn(ListFile, '               ERROR: Translated to ' + XlateString );
   WriteLn(ListFile, '               ERROR: ' + S);
end;


procedure TForm1.LogWarn( S : string );
begin
   if XlateString <> '' then
      WriteLn(ListFile, '               WARN: Translated to ' + XlateString );
   WriteLn(ListFile, '               WARN: ' + S);
end;


procedure TForm1.LogXlate;
begin
   XlateString := words[1] + ' ' + words[2] + ' ' + words[3];
end;


function TForm1.Numericize( S: String ) : integer;
var
   amt, idx, mux, dig, alidx: integer;
   style, c1: Char;
   allowed: string;
   alflag: boolean;
begin
   amt := -99999999;
   style := '?';
   if ( s[1] = '''' ) then
      begin
        style := 'A';
        amt := Ord(s[2]);
      end
   else
   if ( S[1] = '0' ) and ( S[2] = 'b' ) then
      style := 'B'
   else
   if ( S[1] = '0' ) and ( S[2] = 'x' ) then
      style := 'H'
   else
   if ( S[1] = 'B' ) and ( S[2] = '''' ) then
      style := 'B'
   else
   if ( S[1] = '$' ) then
      style := 'H'
   else
   if ( S[1] = 'H' ) and ( S[2] = '''' ) then
      style := 'H'
   else
   if ( ( S[1] >= '0' ) and ( S[1] <= '9' ) ) or ( S[1] = '+' ) then
      begin
         if ( S[Length(S)] = 'H' ) or ( S[Length(S)] = 'h' ) then
            style := 'H'
         else
         if ( S[Length(S)] = 'B' ) or ( S[Length(S)] = 'b' ) then
            style := 'B'
         else
         if style = '?' then
            style := 'D';
      end;
   if style = '?' then
      begin
         idx := 1;
         while ( equ_word[idx] <> ' ' ) and ( idx < 2000 ) do
            begin
               if equ_word[idx] = UpperCase(S) then
                  begin
                     amt := equ_val[idx];
                     idx := 2000;
                  end
               else
                  idx := idx + 1;
            end;
         if idx < 2000 then
            LogError('unrecognized word:' + S);
      end
   else
      if style <> 'A' then
         begin
            if style = 'D' then
               begin
                  allowed := '+Dd''0123456789';
                  mux := 10;
               end
            else
            if style = 'B' then
               begin
                  allowed := '+Bb''01';
                  mux := 2;
               end
            else
            if style = 'H' then
               begin
                  allowed := '+$Hhx''0123456789abcdefABCDEF';
                  mux := 16;
               end;
            amt := 0;
            idx := 1;
            while idx <= Length(S) do
               begin
                  c1 := S[idx];
                  alidx := 1;
                  alflag := false;
                  while alidx <= Length(allowed) do
                     begin
                        if c1 = allowed[alidx] then
                           begin
                              alflag := true;
                              alidx := Length(allowed);
                           end;
                        alidx := alidx + 1;
                     end;
                  if alflag = false then
                     LogError('Invalid character:' + c1 + ' in numeric format, expected one of:' + allowed);
                  if ( c1 >= '0' ) and ( c1 <= '9' ) then
                     begin
                        dig := ord(c1) - ord('0');
                        amt := ( amt * mux ) + dig;
                     end
                  else
                  if ( c1 >= 'A' ) and ( c1 <= 'F' ) and ( mux > 10 ) then
                     begin
                        dig := ord(c1) - ord('A') + 10;
                        amt := ( amt * mux ) + dig;
                     end
                  else
                  if ( c1 >= 'a' ) and ( c1 <= 'f' ) and ( mux > 10 ) then
                     begin
                        dig := ord(c1) - ord('a') + 10;
                        amt := ( amt * mux ) + dig;
                     end;
                  idx := idx + 1;
               end;
         end;
   Numericize := amt;
end;


procedure TForm1.ParseWords( S : string );
var
   parseme: string;
   oneword: string;
   wrdidx, idx: integer;
   inaword, inquotes: boolean;
   quotechar: char;
begin
   wrdidx := 1;
   while wrdidx < 41 do
      begin
         words[wrdidx] := '';
         wrdidx := wrdidx + 1;
      end;
   wrdidx := 1;
   idx := 1;
   inaword := false;
   inquotes := false;
   parseme := S + ';';
   while wrdidx < 41 do
      begin
         if inquotes then
            begin
               if parseme[idx] = quotechar then
                  inquotes := false;
               if parseme[idx] <> '"' then
                  oneword := oneword + parseme[idx];
            end
         else
            begin
               if ( parseme[idx] = ' ' )
                     or ( parseme[idx] = ',' )
                     or ( parseme[idx] = '.' )
                     or ( parseme[idx] = ';' ) then
                  begin
                     if inaword then
                        begin
                           words[wrdidx] := oneword;
                           oneword := '';
                           wrdidx := wrdidx + 1;
                           inaword := false;
                        end;
                     if ( ord(parseme[idx]) = ord(';') ) then
                        wrdidx := 999;
                  end
               else
                  begin
                     if ( parseme[idx] = '''' )
                           or ( parseme[idx] = '"' )
                           or ( parseme[idx] = '\' ) then
                        begin
                           inquotes := true;
                           quotechar := parseme[idx];
                           if quotechar <> '"' then
                              oneword := oneword + parseme[idx];
                        end
                     else
                        oneword := oneword + parseme[idx];
                     inaword := true;
                  end;
            end;
         idx := idx + 1;
      end;
end;

procedure TForm1.WriteEqu;
var
   idx, idx2, swapi: integer;
   swaps, s1: string;
   clean_pass: boolean;
begin
   { first, sort the table }
   clean_pass := false;
   while not clean_pass do
      begin
         clean_pass := true;
         idx := 1;
         idx2 := 2;
         while equ_word[idx2] <> ' ' do
            begin
               if equ_word[idx] > equ_word[idx2] then
                  begin
                     swaps := equ_word[idx2];
                     swapi := equ_val[idx2];
                     equ_word[idx2] := equ_word[idx];
                     equ_val[idx2] := equ_val[idx];
                     equ_word[idx] := swaps;
                     equ_val[idx] := swapi;
                     clean_pass := false;
                  end;
               idx := idx + 1;
               idx2 := idx2 + 1;
            end;
      end;
   idx := 1;
   WriteLn(ListFile, 'EQUATED VALUES:');
   while equ_word[idx] <> ' ' do
      begin
         s1 := Copy( IntToStr(equ_val[idx]) + '       ', 1, 7 );
         WriteLn(ListFile, '     ' + IntToHex(equ_val[idx]) + 'H  '
            + s1 + equ_word[idx] );
         idx := idx + 1;
      end;

end;

procedure TForm1.WriteHex;
var
   idx, gidx, gidx2, count, checksum, val, loc, lo, hi: integer;
   outnums: array [1..8] of integer;
   s1, s2, s3, s4, s5, o1, o2, o3, o4, o5: string;
begin
   { go through the entire memory from 0 to 4999... }
   { gather up to 8 consecutive words per output line }
   idx := 0;
   gidx := 1;
   while idx < 5000 do
      begin
         if ( pm_used[idx] = false ) or ( gidx > 8 ) then
            begin
               if gidx > 1 then
                  begin
                     checksum := ( gidx - 1 ) * 2;
                     s1 := IntToHex(checksum);
                     o1 := ':' + s1[3] + s1[4];
                     lo := loc mod 256;
                     hi := loc div 256;
                     checksum := checksum + lo + hi;
                     o2 := IntToHex(loc);
                     o3 := '00';
                     gidx2 := 1;
                     o4 := '';
                     while gidx2 < gidx do
                        begin
                           val := outnums[gidx2];
                           lo := val mod 256;
                           hi := val div 256;
                           checksum := checksum + lo + hi;
                           s4 := IntToHex(val);
                           o4 := o4 + s4[3] + s4[4] + s4[1] + s4[2];
                           gidx2 := gidx2 + 1;
                        end;
                     s5 := IntToHex(255 - ((checksum + 255) mod 256));
                     o5 := s5[3] + s5[4];
                     gidx := 1;
                     WriteLn( HexFile, o1 + o2 + o3 + o4 + o5);
                  end;
            end;
         if pm_used[idx] then
            begin
               outnums[gidx] := pm_byte[idx];
               if gidx = 1 then
                  loc := idx * 2;
               gidx := gidx + 1;
            end;
         idx := idx + 1;
      end;
   WriteLn( HexFile, ':00000001FF');
end;


{ ***************************** Event response code ***************************** }


procedure TForm1.Button1Click(Sender: TObject);
begin
   if OpenDialog1.Execute then
      begin
         AssembleFile();
         Button2.Enabled := true;
      end;
end;


procedure TForm1.FormCreate(Sender: TObject);
begin
   LoadOpCodes();
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
   AssembleFile();
end;

end.
